/* --- [ tsl2550.c ] -------------------------------------------------------
* Datum       		  : 04.06.2008
* AppNote           : Designer's Notebook Nr. 9 (TAOS)
*							 Simplified TSL2550 Lux Calculation for 
*							 Embedded and Micro Computers (www.taosinc.com)
* Description       : Ermittelt mit vereinfachtem Rechenverfahren aus 
*							 den Messungen der beiden Einzelkanle die Luxzahl
* Tabwidth          : 3
*
****************************************************************************/

#include <avr/io.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <stdint.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#include "TWI_Master.h"	
#include "globals.h"
#include "rs232_2_twi.h"
#include "tsl2550.h"

#ifdef TSL2550
uint16_t tsl2550_count0;
uint16_t tsl2550_count1;
uint8_t  ratio;

uint8_t 	tsl2550_read(void);
uint8_t 	tsl2250_cmd(uint8_t cmd);
uint16_t calculate_lux(void);
uint16_t convert(uint8_t messung);
void wait_10ms(uint8_t t);

// LookupTable im Flash
const uint8_t ratioLUT[129] PROGMEM = {
100, 100, 100, 100, 100, 100, 100, 100,
100, 100, 100, 100, 100, 100,  99,  99,
 99,  99,  99,  99,  99,  99,  99,  99,
 99,  99,  99,  98,  98,  98,  98,  98,
 98,  98,  97,  97,  97,  97,  97,  96,
 96,  96,  96,  95,  95,  95,  94,  94, 
 93,  93,  93,  92,  92,  91,  91,  90,
 89,  89,  88,  87,  87,  86,  85,  84,
 83,  82,  81,  80,  79,  78,  77,  75,
 74,  73,  71,  69,  68,  66,  64,  62,
 60,  58,  56,  54,  52,  49,  47,  44,
 42,  41,  40,  40,  39,  39,  38,  38,
 37,  37,  37,  36,  36,  36,  35,  35,
 35,  35,  34,  34,  34,  34,  33,  33,
 33,  33,  32,  32,  32,  32,  32,  31,
 31,  31,  31,  31,  30,  30,  30,  30,
 30
 };    


// ----------------------------------------------------------------------------
// Konvertiert einen Messwert in einen Helligkeitswert
// (anstelle einer LookUpTable mit 128x uint16_t = 256 Byte)
// Format der Messung: 	Bit 7  			= valid if 1
//								Bit 6, 5, 4		= chord
//								Bit 3, 2, 1, 0	= step			
uint16_t convert(uint8_t messung)
{
uint16_t light, step;					// als uint16_t erforderlich !
uint8_t chord;								// wird nicht grer als 7
												// 0 zurckgeben, 
if (!(messung & 128)) return(0);		// falls das Valid-Bit nicht gesetzt ist

												// den Messwert zur Kontrolle ausgeben
put_string(utoa((messung & 127), msg,10));
put_c('=');

chord = messung & 0x70;					// Bit 6 bis 4 freistellen
chord /= 16;								// nach rechts schieben -> 0b00000111
step = messung & 0x0f; 					// Bit 3 bis 0 freistellen

messung = 1 << chord;					// messung als temporre Variable nutzen
												// (2 ^ chord) zwischenzuspeichern
step *= messung;							// (2 ^ chord) * step
messung--;									// (2 ^ chord - 1)
light = messung * 16;					// 16 *
light += messung / 2;					// (16 + 0.5) * (2 ^ chord -  1)
light += step;

return(light);
}

//-----------------------------------------------------------------------------
// Berechnet den gewichteten Lux-Wert aus den beiden Einzelmessungen
// tsl2550_count0 -> Gesamtspektrum
// tsl2550_count1 -> IR Anteil
uint16_t calculate_lux(void)
{
uint32_t lux = 0;								// Vorgabe fr den Fall dass count1 => count0
uint16_t temp = 0;
uint8_t i, j;

//ratio = 128;									// 128 als Vorgabe
// count0 > 0 und count1 <= count0 - keine Division durch 0 erwnscht
if (tsl2550_count0 && (tsl2550_count1 < tsl2550_count0 ))
	{												
	lux = tsl2550_count1;					// 16bit- Wert zur Multiplikation in 32bit packen
	lux = lux << 7;							// mit 128 multiplizieren
	ratio = (uint8_t) (lux / tsl2550_count0);// und durch count0 dividieren
	temp = tsl2550_count0 - tsl2550_count1;
	
// Die Multiplikation funktioniert nur dann, wenn temp uint32_t ist
// kostet 50 Byte mehr als die Addition unten ...
// lux = temp * pgm_read_byte(ratioLUT[ratio]);  
	j = pgm_read_byte(&ratioLUT[ratio]);// Addition spart RAM

	lux = 0;
	for (i = 0; i < j; i++)
		{
		lux += temp;
		}

	lux /= 256;									// durch 256 dividieren
	}												// wenn count0 = 0 oder count0 < count1
	
if 		(ratio < 20) ratio = 0;			// 0.15 * 128 = 20 - neon
else if 	(ratio < 70) ratio = 1;			// 0.55 * 128 = 70 - shadow
else 		ratio = 2;
	
return((uint16_t) lux);						
}

// ----------------------------------------------------------------------------
// Schickt ein Kommando an den TSL2550
// liefert TRUE wenn erfolgreich, FALSE wenn Fehler
uint8_t tsl2550_cmd(uint8_t cmd)
{
write_buf[0] = TSL_TWI_ADR;
write_buf[1] = cmd;
TWI_Start_Transceiver_With_Data(write_buf, 2);

return (TWI_Get_State_Info() == TWI_NO_STATE);	
}

// ----------------------------------------------------------------------------
// liest ein Byte vom TSL2550
// liefert TRUE wenn erfolgreich, FALSE wenn valid nicht gesetzt, das Byte steht in out_buffer[1]
uint8_t tsl2550_read(void)
{
write_buf[0] = TSL_TWI_ADR + 1;
TWI_Start_Transceiver_With_Data(write_buf, 2);
TWI_Get_Data_From_Transceiver(write_buf, 2);

return(write_buf[1] & 128);						// gibt zurck, ob das valid-Flag gesetzt ist
}

// ----------------------------------------------------------------------------
// liest beide count-Werte aus dem TSL2550 aus und rechnet auf Lux um
void read_lux(void)
{

uint8_t mode = 0;
uint16_t lux;

puts_p(PSTR("TSL2550 "));

if (!(tsl2550_cmd(TSL_POWER_UP)))				// bei Fehler
	{
	err_msg(TWI_Get_State_Info());				// Fehlercode/meldung ausgeben incl. Return
	return;
	}

tsl2550_cmd(TSL_STANDARD_RANGE_MODE);			// erste Messung im Standard-Range
tsl2550_cmd(TSL_READ_CHANNEL_0);

wait_10ms(45);											// feste Wartezeit von 450ms (400ms lt. Manual)

if (tsl2550_read())									// count0 
	{
	if (write_buf[1] > 254)						// wenn der Messwert am Ende des Messbereiches
		{													// dann in den Extended Range schalten und neu messen
		tsl2550_cmd(TSL_EXTENDED_RANGE_MODE);
		mode = 1;										// Flag fr extended range
		tsl2550_cmd(TSL_READ_CHANNEL_0);
		wait_10ms(15);									// feste Wartezeit von 150ms (80ms lt. Manual)
		
		if (!(tsl2550_read()))						// wenn ein Fehler
			{
			puts_p(PSTR("error2"));
			put_cr();
			return;
			}
		}
	}

else
	{
	puts_p(PSTR("error1"));
	put_cr();
	return;
	}

puts_p(PSTR(" Count0: "));
tsl2550_count0 = convert(write_buf[1]);
put_string(ltoa(tsl2550_count0, msg, 10));
	
tsl2550_cmd(TSL_READ_CHANNEL_1);					// danach warten

if (mode) wait_10ms(15); else wait_10ms(45);	// feste Wartezeit von 450ms/150ms
											
if (tsl2550_read())									// count1
	{
	puts_p(PSTR(" Count1: "));
	tsl2550_count1 = convert(write_buf[1]);
	put_string(ltoa(tsl2550_count1, msg, 10));
	}

else
	{
	puts_p(PSTR("error3"));
	put_cr();
	return;
	}
	
lux = calculate_lux();

if (mode)												// wenn im extended range
	{
	puts_p(PSTR(" [ext"));
	lux *= 5;											// im extended range nur 1/5 der Empfindlichkeit
	}

else
	{
	puts_p(PSTR(" [std"));
	}

puts_p(PSTR("-Range] Lux: "));					// Lux
put_string(ultoa(lux, msg, 10));

tsl2550_cmd(TSL_POWER_DOWN);						// power_down

put_c(' ');
if      (ratio == 0) puts_p(PSTR("Leuchtstoff"));	// Auswertung von ratio
else if (ratio == 1) puts_p(PSTR("Tageslicht"));
else                 puts_p(PSTR("Gluehlampe"));
put_cr();

return;
}


// -------------
void wait_10ms(uint8_t t)
{
register uint8_t i;
for(i = 0; i < t; i++) _delay_ms(10);
}
#endif
